<?php

namespace Drupal\Tests\commerce_vado\Kernel\Entity;

use Drupal\commerce_vado\Entity\VadoGroup;
use Drupal\commerce_vado\Entity\VadoGroupItem;
use Drupal\commerce_product\Entity\ProductVariation;
use Drupal\Tests\commerce_cart\Kernel\CartKernelTestBase;

/**
 * Tests the Group entity.
 *
 * @group commerce_vado
 */
class VadoGroupTest extends CartKernelTestBase {

  /**
   * Modules to enable.
   *
   * @var array
   */
  public static $modules = [
    'commerce_vado',
  ];

  /**
   * The vado field manager.
   *
   * @var \Drupal\commerce_vado\VadoFieldManagerInterface
   */
  protected $vadoFieldManager;

  /**
   * A sample user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $user;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('commerce_vado_group_item');
    $this->installEntitySchema('commerce_vado_group');
    $this->installConfig(['commerce_vado']);

    $this->vadoFieldManager = $this->container->get('commerce_vado.field_manager');
    $this->vadoFieldManager->installField('variation_groups', 'default');

    $user = $this->createUser([], ['administer commerce_vado_group']);
    $this->user = $this->reloadEntity($user);
    $this->container->get('current_user')->setAccount($user);

    $this->variation1 = ProductVariation::create([
      'type' => 'default',
      'sku' => strtolower($this->randomMachineName()),
      'title' => $this->randomString(),
      'status' => 0,
    ]);
    $this->variation1->save();

    $this->variation2 = ProductVariation::create([
      'type' => 'default',
      'sku' => strtolower($this->randomMachineName()),
      'title' => $this->randomString(),
      'status' => 1,
    ]);
    $this->variation2->save();

    $this->group_item1 = VadoGroupItem::create();
    $this->group_item1->save();
    $this->group_item1->set('variation', $this->variation1);
    $this->group_item1->save();
    // setItems() and addItem() do not seem to set the group_id backreference.
    $this->group_item1->set('group_id', 1);
    $this->group_item1->save();

    $this->group_item2 = VadoGroupItem::create();
    $this->group_item2->save();
    $this->group_item2->set('variation', $this->variation2);
    $this->group_item2->save();
    // setItems() and addItem() do not seem to set the group_id backreference.
    $this->group_item2->set('group_id', 1);
    $this->group_item2->save();

    $this->group_items = [
      $this->group_item1,
      $this->group_item2,
    ];

    $this->group = VadoGroup::create();
    $this->group->save();
    $this->group->set('title', 'Group 1');
    $this->group->save();
  }

  /**
   * The vado group test.
   */
  public function testVadoGroup() {

    $this->group->setCreatedTime(635879700);
    $this->assertEquals(635879700, $this->group->getCreatedTime());

    $this->group->setRequired(TRUE);
    $this->group->save();
    $this->assertTrue($this->group->isRequired());
    $this->group->setRequired(FALSE);
    $this->group->save();
    $this->assertFalse($this->group->isRequired());

    $this->group->setGroupDiscount(10);
    $this->group->save();
    $this->assertTrue($this->group->hasGroupDiscount());
    $this->assertEquals(10, $this->group->getGroupDiscount());
    $this->assertEquals(.9, $this->group->getGroupDiscountMultiplier());
    $this->assertEquals(.1, $this->group->getGroupDiscountAmountMultiplier());
    $this->group->setGroupDiscount(NULL);
    $this->group->save();
    $this->assertFalse($this->group->hasGroupDiscount());
  }

  /**
   * The vado group item group test.
   */
  public function testVadoGroupItemGroup() {

    $this->group->addItem($this->group_item1);
    $this->group->save();
    $this->assertTrue($this->group->hasItems());
    $this->assertTrue($this->group->hasItem($this->group_item1));
    $this->assertEquals(1, $this->group_item1->id());
    $this->assertEquals(1, $this->group_item1->getGroupId());
    $this->assertNotNull($this->group_item1->getGroup());
    $this->group->RemoveItem($this->group_item1);
    $this->group->save();
    $this->assertFalse($this->group->hasItems());
    $this->assertFalse($this->group->hasItem($this->group_item1));

    $this->group->setItems($this->group_items);
    $this->group->save();
    $this->assertTrue($this->group->hasItems());
    $this->assertTrue($this->group->hasItem($this->group_item1));
    $this->assertTrue($this->group->hasItem($this->group_item2));
    $this->assertEquals(1, $this->group_item1->id());
    $this->assertEquals(2, $this->group_item2->id());
    $this->assertEquals(1, $this->group_item1->getGroupId());
    $this->assertEquals(1, $this->group_item2->getGroupId());
    $this->assertNotNull($this->group_item1->getGroup());
    $this->assertNotNull($this->group_item2->getGroup());
  }

  /**
   * The vado group duplicate test.
   */
  public function testVadoGroupDuplicate() {

    // Set the group values to duplicate.
    $this->group->setRequired(TRUE);
    $this->group->save();
    $this->assertTrue($this->group->isRequired());

    $this->group->setGroupDiscount(10);
    $this->group->save();
    $this->assertTrue($this->group->hasGroupDiscount());
    $this->assertEquals(10, $this->group->getGroupDiscount());
    $this->assertEquals(.9, $this->group->getGroupDiscountMultiplier());
    $this->assertEquals(.1, $this->group->getGroupDiscountAmountMultiplier());

    $this->group->setItems($this->group_items);
    $this->group->save();

    $this->assertTrue($this->group->hasItems());
    $this->assertTrue($this->group->hasItem($this->group_item1));
    $this->assertTrue($this->group->hasItem($this->group_item2));
    $this->assertNotNull($this->group_item1->getGroup());
    $this->assertEquals(1, $this->group_item1->getGroupId());
    $this->assertNotNull($this->group_item2->getGroup());
    $this->assertEquals(1, $this->group_item2->getGroupId());
    $this->assertEquals(1, $this->group_item1->getVariationId());
    $this->assertEquals(2, $this->group_item2->getVariationId());

    // Create the duplicate group, and check it's values.
    $duplicate_group = $this->group->createDuplicate();
    $duplicate_group->save();
    $this->assertEquals('Group 1', $duplicate_group->label());
    $this->assertTrue($duplicate_group->isRequired());
    $this->assertEquals(10, $duplicate_group->getGroupDiscount());
    $this->assertTrue($duplicate_group->hasItems());
    $this->assertFalse($duplicate_group->hasItem($this->group_item1));
    $this->assertFalse($duplicate_group->hasItem($this->group_item2));

    // @todo Figure out a better way to identify the 2 duplicate group items.
    $duplicate_group_items = $duplicate_group->getItems();
    foreach ($duplicate_group_items as $duplicate_group_item) {
      if ($duplicate_group_item->id() == 3) {
        $duplicate_group_item1 = $duplicate_group_item;
      }
      if ($duplicate_group_item->id() == 4) {
        $duplicate_group_item2 = $duplicate_group_item;
      }
    }

    // Check that the duplicate group item id's are new, and successive.
    $this->assertTrue($duplicate_group->hasItem($duplicate_group_item1));
    $this->assertTrue($duplicate_group->hasItem($duplicate_group_item2));
    $this->assertEquals(2, $duplicate_group_item1->getGroupId());
    $this->assertNotNull($duplicate_group_item1->getGroup());
    $this->assertEquals(2, $duplicate_group_item2->getGroupId());
    $this->assertNotNull($duplicate_group_item2->getGroup());
    $this->assertNotEquals(1, $duplicate_group_item1->id());
    $this->assertNotEquals(2, $duplicate_group_item2->id());
    $this->assertEquals(3, $duplicate_group_item1->id());
    $this->assertEquals(4, $duplicate_group_item2->id());
    $this->assertEquals(1, $duplicate_group_item1->getVariationId());
    $this->assertEquals(2, $duplicate_group_item2->getVariationId());

    // Set the discount on original #1 and make sure duplicate #1 is still NULL.
    $this->group_item1->set('group_item_discount', 10);
    $this->group_item1->save();
    $this->assertEquals(10, $this->group_item1->getGroupItemDiscount());
    $this->assertNull($duplicate_group_item1->getGroupItemDiscount());

    // Remove duplicate #1 and make sure original #1 is still there.
    $duplicate_group->RemoveItem($duplicate_group_item1);
    $duplicate_group->save();
    $this->assertTrue($this->group->hasItem($this->group_item1));
    $this->assertFalse($duplicate_group->hasItem($duplicate_group_item1));
    $this->assertTrue($duplicate_group->hasItem($duplicate_group_item2));

    // Set the discount on duplicate #2 and make sure original #2 is still NULL.
    $duplicate_group_item2->set('group_item_discount', 10);
    $duplicate_group_item2->save();
    $this->assertEquals(10, $duplicate_group_item2->getGroupItemDiscount());
    $this->assertNull($this->group_item2->getGroupItemDiscount());

    // Delete the duplicate group and make sure the original group stays.
    $duplicate_group->delete();
    $this->assertTrue($this->group->hasItems());
    $this->assertTrue($this->group->hasItem($this->group_item1));
    $this->assertTrue($this->group->hasItem($this->group_item2));
    $this->assertNotNull($this->group_item1->getGroup());
    $this->assertEquals(1, $this->group_item1->getGroupId());
    $this->assertNotNull($this->group_item2->getGroup());
    $this->assertEquals(1, $this->group_item2->getGroupId());
  }

}
